#include "vtkObjectFactory.h" //for new() macro
#include "vtkInformation.h"
#include "vtkInformationVector.h"
#include "vtkTransform.h"
#include "vtkTransformPolyDataFilter.h"
#include "vtkPolyData.h"
#include "vtkIterativeClosestPointTransform.h"
#include "vtkLandmarkTransform.h"

#include "vtkICPFilter.h"

vtkStandardNewMacro(vtkICPFilter);

//-----------------------------------------------------------------------------
vtkICPFilter::vtkICPFilter()
{
  // Mandatory!
  this->SetNumberOfInputPorts(2);
}


//-----------------------------------------------------------------------------
vtkICPFilter::~vtkICPFilter()
{

}


//----------------------------------------------------------------------------
vtkPolyData *vtkICPFilter::GetTarget()
{
  if (this->GetNumberOfInputConnections(0) < 1)
    {
    return NULL;
    }

  return vtkPolyData::SafeDownCast(this->GetExecutive()->GetInputData(0, 0));
}


//-----------------------------------------------------------------------------
vtkPolyData *vtkICPFilter::GetSource()
{
  if (this->GetNumberOfInputConnections(1) < 1) // because the port is optional
    {
    return 0;
    }
  return static_cast<vtkPolyData *>(this->GetExecutive()->GetInputData(1, 0));
}


//----------------------------------------------------------------------------
void vtkICPFilter::AddSourceConnection(vtkAlgorithmOutput* input)
{
  this->AddInputConnection(1, input);
}


//----------------------------------------------------------------------------
void vtkICPFilter::RemoveAllSources()
{
  this->SetInputConnection(1, 0);
}


//----------------------------------------------------------------------------
int vtkICPFilter::FillInputPortInformation( int port, vtkInformation* info )
{
  if (!this->Superclass::FillInputPortInformation(port, info))
    {
    return 0;
    }
  if ( port == 0 )
    {
    info->Set( vtkAlgorithm::INPUT_REQUIRED_DATA_TYPE(), "vtkPolyData" );
    return 1;
    }
  else if ( port == 1 )
    {
    info->Set( vtkAlgorithm::INPUT_REQUIRED_DATA_TYPE(), "vtkPolyData" );
    return 1;
    }
  return 0;
}


//----------------------------------------------------------------------------
int vtkICPFilter::RequestData(vtkInformation *vtkNotUsed(request),
                              vtkInformationVector **inputVector,
                              vtkInformationVector *outputVector)
{
  // get the info objects
  vtkInformation *inInfo = inputVector[0]->GetInformationObject(0);
  vtkInformation *outInfo = outputVector->GetInformationObject(0);

  // get the input and output
  vtkPolyData *input = vtkPolyData::SafeDownCast(
                                                 inInfo->Get(vtkDataObject::DATA_OBJECT()));
  vtkPolyData *output = vtkPolyData::SafeDownCast(
                                                  outInfo->Get(vtkDataObject::DATA_OBJECT()));

  //ICP Transform
  this->ICPTransform = vtkSmartPointer<vtkIterativeClosestPointTransform>::New();
  this->ICPTransform->SetSource(this->GetSource());
  this->ICPTransform->SetTarget(this->GetTarget());

  //icp->CheckMeanDistanceOn();
  //icp->SetMaximumNumberOfIterations(maxiters);
  //icp->SetThresh(Thresh);
  if(this->bStartByMatchingCentroids)  
	ICPTransform->StartByMatchingCentroidsOn(); //align center of masses as first transformation

  //all points are used if this is not done, the idea is to use a subset of the points if they are very dense
  //icp->SetMaximumNumberOfLandmarks(1000);

  //icp->SetMaximumMeanDistance(accuracy); //the test distance for convergence

  //if you don't do this, it defaults to similarity transform (translation, rotation, and isotropic scaling)
if(this->bRigidBody)
  this->ICPTransform->GetLandmarkTransform()->SetModeToRigidBody();

  //icp->Modified(); //is this necessary?
  this->ICPTransform->Update();

  //bring the source to the target
  vtkSmartPointer<vtkTransformPolyDataFilter> TransformFilter = vtkSmartPointer<vtkTransformPolyDataFilter>::New();
  TransformFilter->SetInputData(this->GetSource());
  TransformFilter->SetTransform(this->ICPTransform);
  TransformFilter->Update();
  output->ShallowCopy(TransformFilter->GetOutput());

  return 1;
}



////////// External Operators /////////////

void vtkICPFilter::PrintSelf(ostream &os, vtkIndent indent)
{
}
